from __future__ import annotations

import logging

logger = logging.getLogger(__name__)


class Allocator:
    _PAGE_SIZE = 0xFF

    def __init__(self, base: int, size: int) -> None:
        self._base = base
        self._size = size

        # allocated blocks: addr -> size
        self._free_blocks: list[tuple[int, int]] = [(self._base, self._base + self._size)]
        self._alloc_size: dict[int, int] = {}

    @property
    def alloc_size(self) -> int:
        return sum(self._alloc_size.values())

    @property
    def alloc_perc(self) -> float:
        return self.alloc_size / self._size

    def _find_free_block(self, size: int) -> int | None:
        return next((i for i, block in enumerate(self._free_blocks) if (block[1] - block[0]) >= size), None)

    def _claim_block(self, block_i: int, size: int) -> int:
        block_start, block_end = self._free_blocks.pop(block_i)

        new_start = block_start + size
        assert new_start <= block_end
        if new_start != block_end:
            self._free_blocks.insert(0, (block_start + size, block_end))

        self._alloc_size[block_start] = size

        return block_start

    def _create_free_block(self, address: int, size: int) -> None:
        blocks = [
            (-1, self._base),
            *self._free_blocks,
            (self._base + self._size, -1),
        ]
        for i, ((_, start), (end, _)) in enumerate(zip(blocks, blocks[1:])):
            if start <= address <= end:  # addr lies in this allocated (non-free) block
                new_block = (address, address + size)

                if self._free_blocks[i - 1][1] == new_block[0]:
                    # can merge with previous free block
                    self._free_blocks[i - 1] = (self._free_blocks[i - 1][0], new_block[1])

                    if new_block[1] == self._free_blocks[i][0]:
                        # can *also* merge with next block
                        self._free_blocks[i - 1] = (self._free_blocks[i - 1][0], self._free_blocks[i][1])
                        del self._free_blocks[i]
                elif new_block[1] == self._free_blocks[i][0]:
                    # can *only* merge with next free block
                    self._free_blocks[i] = (new_block[0], self._free_blocks[i][1])
                else:
                    # cannot merge
                    self._free_blocks.insert(i, new_block)

                return

        logger.warning("Could not free 0x%x: address not allocated?", address)

    def alloc(self, size: int) -> tuple[int, int]:
        length = (size + self._PAGE_SIZE) & ~self._PAGE_SIZE  # Align to pagesize bytes
        block_i = self._find_free_block(length)
        if block_i is None:
            msg = "Cannot alloc more memory: allocator is full!"
            raise RuntimeError(msg)

        address = self._claim_block(block_i, length)

        logger.debug("Allocating %d bytes (align: %d) at 0x%x", size, length, address)
        logger.debug("Allocator base: 0x%x, size: 0x%x", self._base, self._size)
        logger.debug("New alloc size: %x", self.alloc_size)

        return address, length

    def free(self, address: int) -> None:
        size = self._alloc_size.pop(address, None)
        if size is None:
            logger.warning("Tried to free memory at 0x%X, but never allocated", address)
            return

        self._create_free_block(address, size)

        logger.debug("Freed %x bytes at %x, new alloc size: %x", size, address, self.alloc_size)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import anisette

project = "Anisette.py"
release = anisette.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.duration",
    "sphinx.ext.autodoc",
    "sphinx.ext.inheritance_diagram",
    "autoapi.extension",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- AutoAPI Options ---------------------------------------------------------
autoapi_dirs = ["../src/anisette/"]
autoapi_add_toctree_entry = False
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "imported-members",
]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]



from __future__ import annotations

import contextlib
import io
import json
import logging
import os
import tarfile
from dataclasses import dataclass
from pathlib import Path
from typing import IO, BinaryIO, Literal, Union, overload

from typing_extensions import Self

logger = logging.getLogger(__name__)

Directory = dict[str, Union["Directory", bytearray]]

O_RDONLY = 0o0
O_WRONLY = 0o1
O_RDWR = 0o2
O_CREAT = 0o100
O_NOFOLLOW = 0o100000


def split_path(path: str) -> tuple[str, ...]:
    return Path(path).parts


@dataclass(frozen=True)
class StatResult:
    st_mode: int
    st_size: int


class VirtualFileSystem:
    def __init__(self, fs: VirtualFileSystem | None = None) -> None:
        # Share underlying tree if another VFS is provided
        if isinstance(fs, VirtualFileSystem):
            self._tree: Directory = fs.root
        else:
            self._tree = {}

        # fd table: fd -> (path, mode, buffer)
        self._file_handles: dict[int, tuple[str, str, io.BytesIO]] = {}

    @property
    def root(self) -> Directory:
        return self._tree

    # -------- Internal helpers --------
    def _split(self, path: str) -> list[str]:
        parts = [p for p in Path(path).parts if p not in (".", "")]
        if parts and parts[0] == "/":
            parts = parts[1:]
        return parts

    def _get_dir(self, parts: list[str], create: bool = False) -> Directory:
        node: Directory = self._tree
        for part in parts:
            entry = node.get(part)
            if entry is None:
                if not create:
                    raise FileNotFoundError
                sub: Directory = {}
                node[part] = sub
                node = sub
                continue
            if isinstance(entry, bytearray):
                raise NotADirectoryError
            node = entry
        return node

    def _get_parent(self, path: str, create: bool = False) -> tuple[Directory, str]:
        parts = self._split(path)
        if not parts:
            return self._tree, ""
        parent, name = parts[:-1], parts[-1]
        return self._get_dir(parent, create=create), name

    def _get_file(self, path: str) -> bytearray:
        parent, name = self._get_parent(path, create=False)
        entry = parent.get(name)
        if entry is None:
            raise FileNotFoundError
        if isinstance(entry, dict):
            raise IsADirectoryError
        return entry

    def _ensure_file(self, path: str) -> bytearray:
        parent, name = self._get_parent(path, create=True)
        entry = parent.get(name)
        if entry is None or isinstance(entry, dict):
            parent[name] = bytearray()
            return parent[name]  # type: ignore[index]
        return entry

    def read_bytes(self, path: str) -> bytes:
        return bytes(self._get_file(path))

    def write_bytes(self, path: str, data: bytes) -> None:
        file = self._ensure_file(path)
        file[:] = data

    def listdir(self, path: str = ".") -> list[str]:
        parts = self._split(path)
        node = self._get_dir(parts, create=False)
        return list(node.keys())

    def walk(self, top: str = ".") -> list[tuple[str, list[str], list[str]]]:
        parts = self._split(top)
        node = self._get_dir(parts, create=False)
        result: list[tuple[str, list[str], list[str]]] = []

        def _walk(cur_node: Directory, cur_path: str) -> None:
            dirs: list[str] = []
            files: list[str] = []
            for name, entry in cur_node.items():
                if isinstance(entry, dict):
                    dirs.append(name)
                else:
                    files.append(name)
            result.append((cur_path or ".", dirs[:], files[:]))
            for name in dirs:
                sub = cur_node[name]
                assert isinstance(sub, dict)
                sub_path = f"{cur_path}/{name}" if cur_path else name
                _walk(sub, sub_path)

        _walk(node, "/".join(parts))
        return result

    class _VfsOpen:
        def __init__(self, vfs: VirtualFileSystem, path: str, mode: str) -> None:
            self._vfs = vfs
            self._path = path
            self._mode = mode
            self._bin = "b" in mode
            self._write = any(c in mode for c in ("w", "+", "a"))
            if self._write:
                buf = io.BytesIO()
                if "a" in mode:
                    with contextlib.suppress(FileNotFoundError):
                        data = vfs.read_bytes(path)
                        buf.write(data)
                self._buf = buf
            else:
                data = vfs.read_bytes(path)
                self._buf = io.BytesIO(data)

            if self._bin:
                self._stream: IO = self._buf
            else:
                self._stream = io.TextIOWrapper(self._buf, encoding="utf-8")

        def __enter__(self) -> IO:
            return self._stream

        def __exit__(self, exc_type, exc, tb) -> None:  # noqa: ANN001
            try:
                if self._write and exc is None:
                    self._stream.flush()
                    if not isinstance(self._stream, io.TextIOBase):
                        data = self._stream.getbuffer().tobytes()  # type: ignore[assignment]
                    else:
                        # flush encoder and fetch underlying bytes
                        self._stream.flush()
                        data = self._buf.getvalue()
                    self._vfs.write_bytes(self._path, data)
            finally:
                try:
                    self._stream.close()
                finally:
                    self._buf.close()

    def easy_open(self, path: str, mode: str = "r") -> IO:
        # Return a context manager that yields a file-like; callers use `with`.
        return self._VfsOpen(self, path, mode)  # type: ignore[return-value]

    def read(self, fd: int, length: int) -> bytes:
        logger.debug("FS: read %d: %d", fd, length)
        _path, _mode, buf = self._file_handles[fd]
        return buf.read(length)

    def write(self, fd: int, data: bytes) -> None:
        logger.debug("FS: write %d: %s", fd, data.hex())
        _path, _mode, buf = self._file_handles[fd]
        buf.write(data)

    def truncate(self, fd: int, length: int) -> None:
        logger.debug("FS: truncate %d: %d", fd, length)
        _path, _mode, buf = self._file_handles[fd]
        buf.truncate(length)

    def open(self, path: str, o_flag: int) -> int:
        # Determine mode string for our buffer bookkeeping
        mode = "wb+" if (o_flag & O_WRONLY or o_flag & O_RDWR) else "rb"
        if o_flag & O_CREAT and "+" not in mode:
            mode += "+"

        logger.debug("FS: open %s: %s", mode, path)

        # Create buffer for fd
        if "w" in mode or "+" in mode:
            buf = io.BytesIO()
        else:
            data = bytes(self._get_file(path))
            buf = io.BytesIO(data)

        # pick the lowest available fd
        fd = 0
        while fd in self._file_handles:
            fd += 1
        self._file_handles[fd] = (path, mode, buf)
        return fd

    def close(self, fd: int) -> None:
        logger.debug("FS: close %d", fd)

        path, mode, buf = self._file_handles.pop(fd)
        # Commit on write
        if "w" in mode or "+" in mode:
            data = buf.getvalue()
            file = self._ensure_file(path)
            file[:] = data
        buf.close()

    def mkdir(self, path: str) -> None:
        logger.debug("FS: mkdir %s", path)

        try:
            self._get_dir(self._split(path), create=True)
        except NotADirectoryError:
            # trying to mkdir where a file exists
            raise FileExistsError from None

    def stat(self, path_or_fd: str | int) -> StatResult:
        logger.debug("FS: stat %s", path_or_fd)

        if isinstance(path_or_fd, int):  # file descriptor
            _path, _mode, buf = self._file_handles[path_or_fd]
            cur_pos = buf.tell()
            buf.seek(0, os.SEEK_END)
            size = buf.tell()
            buf.seek(cur_pos, os.SEEK_SET)

            return StatResult(
                st_mode=33188,
                st_size=size,
            )

        # path case
        try:
            parent, name = self._get_parent(path_or_fd, create=False)
        except FileNotFoundError:
            raise FileNotFoundError from None
        if name == "":
            # root dir
            return StatResult(st_mode=16877, st_size=4096)
        entry = parent.get(name)
        if entry is None:
            raise FileNotFoundError from None
        if isinstance(entry, dict):
            return StatResult(st_mode=16877, st_size=4096)
        return StatResult(st_mode=33188, st_size=len(entry))


class FSCollection:
    def __init__(self, **filesystems: VirtualFileSystem) -> None:
        self._filesystems = filesystems

    @classmethod
    def load(cls, *files: BinaryIO) -> Self:
        filesystems: dict[str, VirtualFileSystem] = {}
        for f in files:
            with tarfile.open(fileobj=f, mode="r:*") as tf:
                # Read index
                try:
                    member = tf.getmember("fs.json")
                except KeyError:
                    continue
                idx_f = tf.extractfile(member)
                if idx_f is None:
                    continue
                with idx_f:
                    fs_index = json.loads(idx_f.read().decode("utf-8"))

                cls._populate_from_tar(tf, fs_index, filesystems)
        return cls(**filesystems)

    @staticmethod
    def _populate_from_tar(
        tf: tarfile.TarFile,
        fs_index: dict[str, str],
        out: dict[str, VirtualFileSystem],
    ) -> None:
        members = tf.getmembers()
        for name, base in fs_index.items():
            if name in out:
                msg = "Filesystem %s appears in multiple bundles"
                logger.warning(msg, name)

            vfs = VirtualFileSystem()
            base_stripped = base.strip("/")
            prefix = base_stripped + "/"
            for m in members:
                if not m.name.startswith(prefix):
                    continue
                rel = m.name[len(prefix) :]
                if not rel:
                    continue
                if m.isdir():
                    with contextlib.suppress(FileExistsError):
                        vfs.mkdir(rel)
                elif m.isfile():
                    fobj = tf.extractfile(m)
                    if fobj is None:
                        continue
                    with fobj:
                        vfs.write_bytes(rel, fobj.read())
            out[name] = vfs

    def add(self, name: str, fs: VirtualFileSystem) -> None:
        self._filesystems[name] = fs

    def save(self, file: BinaryIO, include: list[str] | None = None, exclude: list[str] | None = None) -> None:
        to_save = set(self._filesystems.keys()) if include is None else set(include)
        if exclude is not None:
            to_save -= set(exclude)

        with tarfile.open(fileobj=file, mode="w:bz2") as tf:
            fs_index: dict[str, str] = {}

            def add_dir(prefix: str, path: str) -> None:
                ti = tarfile.TarInfo(name=f"{prefix}/{path}".rstrip("/"))
                ti.type = tarfile.DIRTYPE
                tf.addfile(ti)

            def add_file(prefix: str, path: str, data: bytes) -> None:
                bio = io.BytesIO(data)
                ti = tarfile.TarInfo(name=f"{prefix}/{path}")
                ti.size = len(data)
                tf.addfile(ti, bio)

            for name in to_save:
                logger.debug("Saving %s to FS bundle", name)
                fs = self._filesystems[name]
                base = f"./{name}"
                fs_index[name] = base

                # ensure root dir
                add_dir(base, ".")
                for dirpath, _dirnames, filenames in fs.walk("."):
                    # add directories explicitly
                    if dirpath != ".":
                        add_dir(base, dirpath)
                    for filename in filenames:
                        rel = filename if dirpath == "." else f"{dirpath}/{filename}"
                        add_file(base, rel, fs.read_bytes(rel))

            # write fs.json
            idx = json.dumps(fs_index).encode("utf-8")
            ti = tarfile.TarInfo(name="fs.json")
            ti.size = len(idx)
            tf.addfile(ti, io.BytesIO(idx))

    @overload
    def get(self, fs_name: str) -> VirtualFileSystem: ...

    @overload
    def get(self, fs_name: str, create_if_missing: Literal[True]) -> VirtualFileSystem: ...

    @overload
    def get(self, fs_name: str, create_if_missing: Literal[False]) -> VirtualFileSystem | None: ...

    def get(self, fs_name: str, create_if_missing: bool = True) -> VirtualFileSystem | None:
        if fs_name in self._filesystems:
            logger.debug("Get FS from collection: %s", fs_name)
            return self._filesystems[fs_name]

        if not create_if_missing:
            return None

        logger.debug("Create new VFS: %s", fs_name)
        fs = VirtualFileSystem()
        self._filesystems[fs_name] = fs
        return fs



from __future__ import annotations

import logging
from typing import BinaryIO, Callable

from typing_extensions import Self

from ._adi import ADI
from ._device import AnisetteDeviceConfig, Device
from ._fs import FSCollection, VirtualFileSystem
from ._library import LibraryStore
from ._session import ProvisioningSession

logger = logging.getLogger(__name__)


class AnisetteProvider:
    def __init__(
        self,
        fs_collection: FSCollection,
        fs_fallback: Callable[[], VirtualFileSystem],
        default_device_config: AnisetteDeviceConfig | None,
    ) -> None:
        self._fs_collection = fs_collection
        self._fs_fallback = fs_fallback
        self._default_device_config = default_device_config or AnisetteDeviceConfig.default()

        self._lib_store: LibraryStore | None = None
        self._device: Device | None = None
        self._adi: ADI | None = None
        self._provisioning_session: ProvisioningSession | None = None

    @classmethod
    def load(
        cls,
        *files: BinaryIO,
        fs_fallback: Callable[[], VirtualFileSystem],
        default_device_config: AnisetteDeviceConfig | None = None,
    ) -> Self:
        provider = cls(FSCollection.load(*files), fs_fallback, default_device_config)
        assert provider.library_store is not None  # verify that library store exists
        return provider

    def save(self, file: BinaryIO, include: list[str] | None = None, exclude: list[str] | None = None) -> None:
        return self._fs_collection.save(file, include, exclude)

    @property
    def library_store(self) -> LibraryStore:
        if self._lib_store is None:
            lib_fs = self._fs_collection.get("libs", create_if_missing=False)
            if lib_fs is None:
                lib_fs = self._fs_fallback()
                self._fs_collection.add("libs", lib_fs)
            self._lib_store = LibraryStore(lib_fs)
        return self._lib_store

    @property
    def device(self) -> Device:
        if self._device is None:
            device_fs = self._fs_collection.get("device")
            self._device = Device(device_fs, self._default_device_config)

        return self._device

    @property
    def adi(self) -> ADI:
        if self._adi and any(usage >= 0.5 for usage in self._adi.alloc_stats):
            logger.warning("Detected memory leak, restarting VM. Next data fetch may take slightly longer.")
            self._adi = None

        if self._adi is None:
            adi_fs = self._fs_collection.get("adi")
            self._adi = ADI(adi_fs, self.library_store, self.device.adi_identifier)

            if self._provisioning_session is not None:
                self._provisioning_session.adi = self._adi

        return self._adi

    @property
    def provisioning_session(self) -> ProvisioningSession:
        if self._provisioning_session is None:
            cache_fs = self._fs_collection.get("cache")
            self._provisioning_session = ProvisioningSession(cache_fs, self.adi, self.device)

        return self._provisioning_session



import logging
import time

import psutil

from anisette import Anisette

logging.basicConfig(level=logging.DEBUG)


def test_memory():
    process = psutil.Process()

    ani = Anisette.init()
    ani.provision()

    start_time = time.time()
    for i in range(2001):
        ani.get_data()

        if i % 100 == 0:
            alloc_stats = ani._ani_provider.adi._vm.alloc_stats
            mem_usage = process.memory_info().rss / (1024 * 1024)

            print(f"{i} / {time.time() - start_time:.2f}s - {alloc_stats} / {mem_usage} MB")



#!/usr/bin/env python3

"""Find supported python versions for this package."""

import json
from collections.abc import Generator
from itertools import count
from pathlib import Path

import tomli
from packaging.specifiers import SpecifierSet
from packaging.version import Version

MAX_VERSION = 13  # 3.13


def get_python_versions() -> Generator[str, None, None]:
    """Get all python versions this package is compatible with."""
    with Path("pyproject.toml").open("rb") as f:
        pyproject_data = tomli.load(f)

    specifier = SpecifierSet(pyproject_data["project"]["requires-python"])

    below_spec = True
    for v_minor in count():
        version = Version(f"3.{v_minor}")

        # above max version: return
        if v_minor > MAX_VERSION:
            return

        # in specifier: yield
        if version in specifier:
            below_spec = False
            yield str(version)
            continue

        # below specifier: skip
        if below_spec:
            continue

        # above specifier: return
        return


print(json.dumps(list(get_python_versions())))



from __future__ import annotations

import base64
import json
import logging
import plistlib
import ssl
from datetime import datetime
from pathlib import Path
from typing import IO, TYPE_CHECKING

import urllib3

if TYPE_CHECKING:
    from ._adi import ADI
    from ._device import Device
    from ._fs import VirtualFileSystem

ENABLE_CACHE = False

logger = logging.getLogger(__name__)


def get_ssl_context() -> ssl.SSLContext:
    pem_path = Path(__file__).parent / "apple-root.pem"
    return ssl.create_default_context(cafile=pem_path)


def time() -> str:
    # Replaces Clock.currTime().stripMilliseconds().toISOExtString()
    return datetime.now().astimezone().replace(microsecond=0).isoformat()


class ProvisioningSession:
    def __init__(self, fs: VirtualFileSystem, adi: ADI, device: Device) -> None:
        self._fs = fs
        self._adi = adi

        self._http = urllib3.PoolManager(ssl_context=get_ssl_context())

        self.__urlBag = {}

        self.__headers = {
            "User-Agent": "akd/1.0 CFNetwork/1404.0.5 Darwin/22.3.0",
            # they are somehow not using the plist content-type in AuthKit
            "Content-Type": "application/x-www-form-urlencoded",
            "Connection": "keep-alive",
            "X-Mme-Device-Id": device.unique_device_identifier,
            # on macOS, MMe for the Client-Info header is written with 2 caps, while on Windows it is Mme...
            # and HTTP headers are supposed to be case-insensitive in the HTTP spec...
            "X-MMe-Client-Info": device.server_friendly_description,
            "X-Apple-I-MD-LU": device.local_user_uuid,
            # "X-Apple-I-MLB": device.logicBoardSerialNumber, // 17 letters, uppercase in Apple's base 34
            # "X-Apple-I-ROM": device.romAddress, // 6 bytes, lowercase hexadecimal
            # "X-Apple-I-SRL-NO": device.machineSerialNumber, // 12 letters, uppercase
            # different apps can be used, I already saw fmfd and Setup here
            # and Reprovision uses Xcode in some requests, so maybe it is possible here too.
            "X-Apple-Client-App-Name": "Setup",
        }

    @property
    def adi(self) -> ADI:
        return self._adi

    @adi.setter
    def adi(self, adi: ADI) -> None:
        logger.debug("Attached new ADI to ProvisioningSession")
        self._adi = adi

    def _open_cache(self, key: str, mode: str) -> IO:
        return self._fs.easy_open(key, mode)

    def _request(
        self,
        method: str,
        url: str,
        extra_headers: dict[str, str],
        data: str | None = None,
        cache_key: str | None = None,
    ) -> bytes:
        if ENABLE_CACHE and cache_key is not None:
            with self._open_cache(cache_key, "rb") as f:
                return f.read()

        headers = self.__headers | extra_headers
        response = self._http.request(method, url, body=data, headers=headers, timeout=5.0)
        resp_data = response.data
        if cache_key is not None:
            with self._open_cache(f"{cache_key}-head", "w") as f:
                json.dump(headers, f, indent=2)
            with self._open_cache(cache_key, "wb") as f:
                f.write(resp_data)
        return resp_data

    def _get(self, url: str, extra_headers: dict[str, str], cache_key: str | None = None) -> bytes:
        return self._request("GET", url, extra_headers, cache_key=cache_key)

    def _post(self, url: str, data: str, extra_headers: dict[str, str], cache_key: str | None = None) -> bytes:
        return self._request("POST", url, extra_headers, data=data, cache_key=cache_key)

    def load_url_bag(self) -> None:
        content = self._get(
            "https://gsa.apple.com/grandslam/GsService2/lookup",
            {},
            "lookup.xml",
        )
        plist = plistlib.loads(content)
        urls = plist["urls"]
        for url_name, url in urls.items():
            self.__urlBag[url_name] = url

    def provision(self, ds_id: int) -> None:
        logger.debug("ProvisioningSession.provision")
        # FIXME: !!!

        if len(self.__urlBag) == 0:
            self.load_url_bag()

        extra_headers = {
            "X-Apple-I-Client-Time": time(),
        }
        start_provisioning_plist = self._post(
            self.__urlBag["midStartProvisioning"],
            """<?xml version=\"1.0\" encoding=\"UTF-8\"?>
                                     <!DOCTYPE plist PUBLIC \"-//Apple//DTD PLIST 1.0//EN\" \"http://www.apple.com/DTDs/PropertyList-1.0.dtd\">
                                     <plist version=\"1.0\">
                                     <dict>
                                     \t<key>Header</key>
                                     \t<dict/>
                                     \t<key>Request</key>
                                     \t<dict/>
                                     </dict>
                                     </plist>""",
            extra_headers,
            "midStartProvisioning.xml",
        )

        spim_plist = plistlib.loads(start_provisioning_plist)
        spim_response = spim_plist["Response"]
        spim_str = spim_response["spim"]
        logger.debug(spim_str)

        spim = base64.b64decode(spim_str)

        cpim = self._adi.start_provisioning(ds_id, spim)
        # FIXME: scope (failure) try { adi.destroyProvisioning(cpim.session); } catch(Throwable) {}

        logger.debug("cpim: %s", cpim.cpim)

        extra_headers = {
            "X-Apple-I-Client-Time": time(),
        }
        end_provisioning_plist = self._post(
            self.__urlBag["midFinishProvisioning"],
            """<?xml version=\"1.0\" encoding=\"UTF-8\"?>
<!DOCTYPE plist PUBLIC \"-//Apple//DTD PLIST 1.0//EN\" \"http://www.apple.com/DTDs/PropertyList-1.0.dtd\">
<plist version=\"1.0\">
<dict>
\t<key>Header</key>
\t<dict/>
\t<key>Request</key>
\t<dict>
\t\t<key>cpim</key>
\t\t<string>{}</string>
\t</dict>
</dict>
</plist>""".format(base64.b64encode(cpim.cpim).decode("utf-8")),
            extra_headers,
            "midFinishProvisioning.xml",
        )

        plist = plistlib.loads(end_provisioning_plist)
        spim_response = plist["Response"]

        # scope ulong routingInformation;
        # routingInformation = to!ulong(spimResponse["X-Apple-I-MD-RINFO"])
        persistent_token_metadata = base64.b64decode(spim_response["ptm"])
        trust_key = base64.b64decode(spim_response["tk"])

        self._adi.end_provisioning(cpim.session, persistent_token_metadata, trust_key)



"""Anisette provider in a Python package."""

from __future__ import annotations

import base64
import locale
import logging
from contextlib import ExitStack
from ctypes import c_ulonglong
from datetime import datetime
from typing import TYPE_CHECKING, BinaryIO, TypedDict

from typing_extensions import Self

from ._ani_provider import AnisetteProvider
from ._fs import FSCollection
from ._library import LibraryStore
from ._util import open_file

if TYPE_CHECKING:
    from pathlib import Path

    from ._device import AnisetteDeviceConfig


DEFAULT_LIBS_URL = "https://anisette.dl.mikealmel.ooo/libs?arch=arm64-v8a"

logger = logging.getLogger(__name__)


AnisetteHeaders = TypedDict(
    "AnisetteHeaders",
    {
        "X-Apple-I-Client-Time": str,
        "X-Apple-I-MD": str,
        "X-Apple-I-MD-LU": str,
        "X-Apple-I-MD-M": str,
        "X-Apple-I-MD-RINFO": str,
        "X-Apple-I-SRL-NO": str,
        "X-Apple-I-TimeZone": str,
        "X-Apple-Locale": str,
        "X-MMe-Client-Info": str,
        "X-Mme-Device-Id": str,
    },
)


def _get_libs(file: BinaryIO | str | Path | None = None) -> LibraryStore:
    file = file or DEFAULT_LIBS_URL

    with open_file(file, "rb") as f:
        return LibraryStore.from_file(f)


class Anisette:
    """
    The main Anisette provider class.

    This is the main Anisette provider class, which provides the user-facing functionality of this package.
    Each instance of :class:`Anisette` represents a single Anisette session.

    This class should not be instantiated directly through its __init__ method.
    Instead, you should use :meth:`Anisette.init` or :meth:`Anisette.load` depending on your use case.
    """

    def __init__(self, ani_provider: AnisetteProvider) -> None:
        """
        Init.

        :meta private:
        """
        self._ani_provider = ani_provider

        self._ds_id = c_ulonglong(-2).value

    @property
    def is_provisioned(self) -> bool:
        """Whether this Anisette session has been provisioned yet or not."""
        return self._ani_provider.adi.is_machine_provisioned(self._ds_id)

    @classmethod
    def init(
        cls,
        file: BinaryIO | str | Path | None = None,
        default_device_config: AnisetteDeviceConfig | None = None,
    ) -> Self:
        """
        Initialize a new Anisette session from an Apple Music APK or Anisette.py library file.

        The file type will be detected automatically. If :param:`file` is not provided, a library
        bundle will be downloaded automatically. This file is usually a few megabytes large.

        :param file: A file, path or URL to a library file or Apple Music APK.
        :type file: BinaryIO, str, Path, None
        :return: An instance of :class:`Anisette`.
        :rtype: :class:`Anisette`
        """
        ani_provider = AnisetteProvider(
            FSCollection(),
            lambda: _get_libs(file),
            default_device_config,
        )
        return cls(ani_provider)

    @classmethod
    def load(cls, *files: BinaryIO | str | Path, default_device_config: AnisetteDeviceConfig | None = None) -> Self:
        """
        Load a previously-initialized Anisette session.

        Required files can be obtained using the :meth:`Anisette.save_provisioning`, :meth:`Anisette.save_libs`
        and/or :meth:`Anisette.save_all` methods.

        :param files: File objects or paths that together form the provider's virtual file system.
        :type files: BinaryIO, str, Path
        :return: An instance of :class:`Anisette`.
        :rtype: :class:`Anisette`
        """
        with ExitStack() as stack:
            file_objs = [stack.enter_context(open_file(f, "rb")) for f in files]
            ani_provider = AnisetteProvider.load(
                *file_objs,
                fs_fallback=lambda: _get_libs(),
                default_device_config=default_device_config,
            )

        return cls(ani_provider)

    def save_provisioning(self, file: BinaryIO | str | Path) -> None:
        """
        Save provisioning data of this Anisette session to a file.

        The size of this file is usually in the order of kilobytes.

        Saving provisioning data is required if you want to re-use this session at a later time.

        A session may be reconstructed from saved data using the :meth:`Anisette.load` method.

        The advantage of using this method over :meth:`Anisette.save_all` is that it results in less overall disk usage
        when saving many sessions, since library data can be saved separately and may be re-used across sessions.

        :param file: The file or path to save provisioning data to.
        :type file: BinaryIO, str, Path
        """
        self.provision()

        with open_file(file, "wb+") as f:
            self._ani_provider.save(f, exclude=["libs"])

    def save_libs(self, file: BinaryIO | str | Path) -> None:
        """
        Save library data to a file.

        The size of this file is usually in the order of megabytes.

        Library data is session-agnostic and may be used in as many sessions as you wish.
        It can also be used to initialize a new session, without requiring the full Apple Music APK.

        The advantage of using this method over :meth:`Anisette.save_all` is that it results in less overall disk usage
        when saving many sessions, since library data can be saved separately and may be re-used across sessions.

        :param file: The file or path to save library data to.
        :type file: BinaryIO, str, Path
        """
        # force fetch of library store to make sure it exists when saving
        _ = self._ani_provider.library_store

        with open_file(file, "wb+") as f:
            self._ani_provider.save(f, include=["libs"])

    def save_all(self, file: BinaryIO | str | Path) -> None:
        """
        Save a complete copy of this Anisette session to a file.

        The size of this file is usually in the order of megabytes.

        Saving session data is required if you want to re-use this session at a later time.

        A session may be reconstructed from saved data using the :meth:`Anisette.load` method.

        The advantage of using this method over :meth:`Anisette.save_provisioning` and :meth:`Anisette.save_libs`
        is that it is easier to use, since all information to reconstruct the session is contained in a single file.

        :param file: The file or path to save session data to.
        :type file: BinaryIO, str, Path
        """
        with open_file(file, "wb+") as f:
            self._ani_provider.save(f)

    def provision(self) -> None:
        """
        Provision the virtual device, if it has not been provisioned yet.

        In most cases it is not necessary to manually use this method, since :meth:`Anisette.get_data`
        will call it implicitly.
        """
        if not self.is_provisioned:
            logger.info("Provisioning...")
            self._ani_provider.provisioning_session.provision(self._ds_id)

    def get_data(self) -> AnisetteHeaders:
        """
        Obtain Anisette headers for this session.

        :return: Anisette headers that may be used for authentication purposes.
        """
        self.provision()
        otp = self._ani_provider.adi.request_otp(self._ds_id)
        device = self._ani_provider.device

        return {
            "X-Apple-I-Client-Time": datetime.now().astimezone().replace(microsecond=0).isoformat() + "Z",
            "X-Apple-I-MD": base64.b64encode(bytes(otp.otp)).decode(),
            "X-Apple-I-MD-LU": base64.b64encode(str(device.local_user_uuid).encode()).decode(),
            "X-Apple-I-MD-M": base64.b64encode(bytes(otp.machine_id)).decode(),
            "X-Apple-I-MD-RINFO": "17106176",
            "X-Apple-I-SRL-NO": "0",
            "X-Apple-I-TimeZone": str(datetime.now().astimezone().tzinfo),
            "X-Apple-Locale": locale.getlocale()[0] or "en_US",
            "X-MMe-Client-Info": device.server_friendly_description,
            "X-Mme-Device-Id": device.unique_device_identifier,
        }



"""Anisette provider in a Python package."""

from importlib.metadata import version

from ._device import AnisetteDeviceConfig
from .anisette import Anisette, AnisetteHeaders

__version__ = version("anisette")

__all__ = ("Anisette", "AnisetteDeviceConfig", "AnisetteHeaders")



from __future__ import annotations

import logging
import os
import platform
import re
from contextlib import contextmanager
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO, Literal

import certifi
import urllib3

if TYPE_CHECKING:
    from collections.abc import Iterator


URL_REGEX = re.compile(r"^https?://[^\s/$.?#].\S*$")

logger = logging.getLogger(__name__)


def u_to_s32(value: int) -> int:
    b = int.to_bytes(value, 4, "little", signed=False)
    return int.from_bytes(b, "little", signed=True)


def u_to_s64(value: int) -> int:
    b = int.to_bytes(value, 8, "little", signed=False)
    return int.from_bytes(b, "little", signed=True)


def s_to_u32(value: int) -> int:
    b = int.to_bytes(value, 4, "little", signed=True)
    return int.from_bytes(b, "little", signed=False)


def s_to_u64(value: int) -> int:
    b = int.to_bytes(value, 8, "little", signed=True)
    return int.from_bytes(b, "little", signed=False)


@contextmanager
def open_file(fp: BinaryIO | str | Path, mode: Literal["rb", "wb+"] = "rb") -> Iterator[BinaryIO]:
    if isinstance(fp, str):
        if URL_REGEX.match(fp):
            with urllib3.PoolManager(ca_certs=certifi.where()) as http:
                r = http.request("GET", fp)
            fp = BytesIO(r.data)
        else:
            fp = Path(fp)

    if isinstance(fp, Path):
        file = fp.open(mode)
        do_close = True
    elif isinstance(fp, (BinaryIO, BytesIO)):
        file = fp
        file.seek(0)
        do_close = False
    else:
        raise TypeError

    yield file

    if do_close:
        file.close()


def get_config_dir(dir_name: str) -> Path | None:
    plat = platform.system()
    if plat == "Windows":
        path_str = os.getenv("LOCALAPPDATA")
    elif plat in ("Linux", "Darwin"):
        path_str = os.getenv("XDG_CONFIG_HOME")
        if path_str is None:
            home = os.getenv("HOME")
            if home is None:
                logger.info("Could not determine home directory")
                return None
            subpath = os.path.join(home, ".config" if plat == "Linux" else "Library/Preferences")  # noqa: PTH118
            path_str = os.getenv("XDG_CONFIG_HOME", subpath)
    else:
        logger.info("Platform unsupported: %s", plat)
        return None

    if path_str is None:
        logger.info("Could not determine config directory")
        return None

    path = Path(path_str) / dir_name
    path.mkdir(parents=True, exist_ok=True)
    return path



"""CLI functionality."""

# ruff: noqa: T201

from __future__ import annotations

import hashlib
import json
import logging
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from typing import TYPE_CHECKING, Annotated, Callable, override

try:
    import typer
    from rich.console import Console
    from rich.table import Table
except ImportError:
    msg = "Failed to find CLI dependencies. Install the 'anisette[cli]' package if you require CLI support."
    raise ImportError(msg) from None

from ._util import get_config_dir
from .anisette import Anisette

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)
app = typer.Typer(no_args_is_help=True)
console = Console()


class _AniError(Exception):
    pass


class _SessionManager:
    def __init__(self, conf_dir: Path | None = None) -> None:
        self._conf_dir: Path | None = conf_dir

    @property
    def can_save(self) -> bool:
        return self.config_dir is not None

    @property
    def config_dir(self) -> Path | None:
        if self._conf_dir is not None:
            return self._conf_dir

        self._conf_dir = get_config_dir("anisette-py")
        if self._conf_dir is None:
            logger.warning("Could not find user config directory")
            return None
        return self._conf_dir

    @property
    def libs_path(self) -> Path | None:
        if self.config_dir is None:
            return None
        return self.config_dir / "libs.bin"

    def _get_prov_path(self, name: str) -> Path:
        assert self.config_dir is not None

        return self.config_dir / f"{name}.prov"

    def save(self, session: Anisette, name: str) -> None:
        assert self.config_dir is not None

        prov_path = self._get_prov_path(name)
        session.save_provisioning(prov_path)

    def get_hash(self, name: str) -> str:
        assert self.config_dir is not None

        prov_path = self._get_prov_path(name)
        if not prov_path.exists():
            msg = f"Session does not exist: '{name}'"
            raise _AniError(msg)

        with prov_path.open("rb") as f:
            return hashlib.sha256(f.read()).hexdigest()[:10]

    def new(self, name: str) -> Anisette:
        assert self.config_dir is not None
        assert self.libs_path is not None

        prov_path = self._get_prov_path(name)
        if prov_path.exists():
            msg = f"Session with name '{name}' already exists"
            raise _AniError(msg)

        if not self.libs_path.exists():
            session = Anisette.init()
            session.save_libs(self.libs_path)
        else:
            session = Anisette.init(self.libs_path)
        self.save(session, name)

        return session

    def remove(self, name: str) -> None:
        assert self.config_dir is not None

        self._get_prov_path(name).unlink(missing_ok=True)

    def get(self, name: str) -> Anisette:
        assert self.config_dir is not None
        assert self.libs_path is not None

        if not self.libs_path.exists():
            msg = "Libraries are not available"
            raise _AniError(msg)

        prov_path = self._get_prov_path(name)
        if not prov_path.exists():
            msg = f"Session with name '{name}' does not exist"
            raise _AniError(msg)

        return Anisette.load(self.libs_path, prov_path)

    def list(self) -> list[str]:
        assert self.config_dir is not None

        return [path.stem for path in self.config_dir.glob("*.prov")]


class _HttpRequestHandler(SimpleHTTPRequestHandler):
    def __init__(self, ani: Anisette, callback: Callable, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        self._ani = ani
        self._callback = callback

        super().__init__(*args, **kwargs)

    @override
    def do_GET(self) -> None:
        data = self._ani.get_data()

        self.send_response(200)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(data).encode())

        self._callback()


@app.command()
def new(name: Annotated[str, typer.Argument(help="The name of the new session")] = "default") -> None:
    """Create a new Anisette session."""
    sessions = _SessionManager()
    if not sessions.can_save:
        print("Unable to figure out a config directory to store new sessions")
        raise typer.Exit(code=1)

    try:
        sessions.new(name)
    except _AniError as e:
        print(str(e))
        raise typer.Abort from None
    print(f"Successfully created new session: '{name}'")


@app.command()
def remove(name: Annotated[str, typer.Argument(help="The name of the saved session to remove")] = "default") -> None:
    """Remove a saved Anisette session."""
    sessions = _SessionManager()
    if not sessions.can_save:
        print("Unable to figure out a config directory to retrieve sessions from")
        raise typer.Exit(code=1)

    try:
        sessions.remove(name)
    except _AniError as e:
        print(str(e))
        raise typer.Abort from None
    print(f"Successfully destroyed session: '{name}'")


@app.command()
def get(name: Annotated[str, typer.Argument(help="The name of the saved session")] = "default") -> None:
    """Get Anisette data for a saved session."""
    sessions = _SessionManager()
    if not sessions.can_save:
        print("Unable to figure out a config directory to retrieve sessions from")
        raise typer.Exit(code=1)

    try:
        ani = sessions.get(name)
    except _AniError as e:
        print(str(e))
        raise typer.Abort from None
    data = ani.get_data()
    sessions.save(ani, name)

    print(json.dumps(data, indent=2))


@app.command(name="list")
def list_() -> None:
    """List Anisette sessions."""
    sessions = _SessionManager()
    if not sessions.can_save:
        print("Unable to figure out a config directory to retrieve sessions from")
        raise typer.Exit(code=1)

    table = Table("Session name", "Revision")
    for name in sessions.list():
        digest = sessions.get_hash(name)
        table.add_row(name, digest)
    console.print(table)


@app.command()
def serve(
    name: Annotated[str, typer.Argument(help="The name of the saved session")] = "default",
    host: Annotated[str, typer.Option(help="Host to run the server on")] = "localhost",
    port: Annotated[int, typer.Option(help="Port to run the server on")] = 6969,
) -> None:
    """Serve Anisette data for a saved session."""
    sessions = _SessionManager()
    if not sessions.can_save:
        print("Unable to figure out a config directory to retrieve sessions from")
        raise typer.Exit(code=1)

    try:
        ani = sessions.get(name)
    except _AniError as e:
        print(str(e))
        raise typer.Abort from None

    server = ThreadingHTTPServer(
        (host, port),
        lambda *args, **kwargs: _HttpRequestHandler(
            ani,
            (lambda: sessions.save(ani, name)),
            *args,
            **kwargs,
        ),
    )

    print(f"Starting server on {host}:{port}")
    print("Press CTRL+C to exit")
    print()

    try:
        server.serve_forever()
    except KeyboardInterrupt:
        print()
        print("Stopping server")


if __name__ == "__main__":
    app()



from __future__ import annotations

import logging
import math
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

from unicorn import UC_MEM_FETCH_UNMAPPED, UC_MEM_WRITE_UNMAPPED
from unicorn.arm64_const import (
    UC_ARM64_REG_FP,
    UC_ARM64_REG_W13,
    UC_ARM64_REG_W14,
    UC_ARM64_REG_W15,
    UC_ARM64_REG_X0,
    UC_ARM64_REG_X1,
    UC_ARM64_REG_X2,
)

from ._structs import c_stat, c_timeval
from ._util import s_to_u64

if TYPE_CHECKING:
    from ._fs import VirtualFileSystem
    from ._vm import VM

logger = logging.getLogger(__name__)

IMPORT_ADDRESS = 0xA0000000
IMPORT_SIZE = 0x1000


@dataclass()
class HookContext:
    vm: VM
    fs: VirtualFileSystem


def _hook_empty_stub(ctx: HookContext, orig_name: str | None = None) -> None:
    if orig_name is not None:
        logger.debug("%s(???) - (empty stubbed)", orig_name)
    ctx.vm.reg_write(UC_ARM64_REG_X0, 0)


def _hook_malloc(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    logger.debug("malloc(0x%X)", x0)
    x0 = ctx.vm.malloc(x0)
    ctx.vm.reg_write(UC_ARM64_REG_X0, x0)


def _hook_free(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    logger.debug("free(0x%X)", x0)

    ctx.vm.free(x0)

    ctx.vm.reg_write(UC_ARM64_REG_X0, 0)


def _hook_strncpy(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)
    x2 = ctx.vm.reg_read(UC_ARM64_REG_X2)

    p_dst = x0
    p_src = x1
    _len = x2

    src = ctx.vm.read_cstr(p_src)
    if len(src) > _len:
        msg = f"Read data too long: {len(src)} > {_len}"
        raise RuntimeError(msg)

    padding_size = _len - len(src)
    data = src + b"\x00" * padding_size

    ctx.vm.mem_write(p_dst, data)

    ctx.vm.reg_write(UC_ARM64_REG_X0, p_dst)


def _hook_mkdir(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)

    path = ctx.vm.read_cstr(x0).decode("utf-8")
    mode = x1

    logger.debug("mkdir('%s', %s)", path, oct(mode))

    assert path in [
        "./anisette",
    ]
    assert mode == 0o777
    ctx.fs.mkdir(path)  # FIXME: mode?

    ctx.vm.reg_write(UC_ARM64_REG_X0, 0)


def _hook_umask(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)

    cmask = x0

    cmask = 0o777

    ctx.vm.reg_write(UC_ARM64_REG_X0, cmask)


def _hook_chmod(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)

    path = ctx.vm.read_cstr(x0).decode("utf-8")
    mode = x1

    logger.debug("chmod('%s', %s)", path, oct(mode))

    ctx.vm.reg_write(UC_ARM64_REG_X0, 0)


def _handle_stat(ctx: HookContext, path_or_fd: str | int, buf: int) -> None:
    try:
        stat_result = ctx.fs.stat(path_or_fd)
        # print(statResult)
    except FileNotFoundError:
        logger.debug("Unable to stat '%s'", path_or_fd)
        ctx.vm.reg_write(UC_ARM64_REG_X0, s_to_u64(-1))
        ctx.vm.set_errno(2)  # ENOENT
        return

    stat = c_stat(
        st_dev=0,
        st_ino=0,
        st_mode=stat_result.st_mode,
        # ...
        st_size=stat_result.st_size,
        st_blksize=512,
        st_blocks=(stat_result.st_size + 511) // 512,
        # ...s
    )
    stat.__byte = stat_result.st_size  # noqa: SLF001
    stat_bytes = bytes(stat)
    # print(statBytes.hex(), len(statBytes))

    # logger.debug("%s %s %s", stat_result.st_size, stat_result.st_blksize, stat_result.st_blocks)
    logger.debug("%s %s %s", stat.st_size, stat.st_blksize, stat.st_blocks)

    logger.debug("0x%X = %d", stat_result.st_mode, stat_result.st_mode)
    stat_bytes = b"".join(
        [
            bytes.fromhex(
                "00000000"
                "00000000"  # st_dev
                "00000000"
                "00000000",
            )  # st_ino
            + int.to_bytes(stat_result.st_mode, 4, "little")  # st_mode
            + bytes.fromhex(
                "00000000"  # st_nlink
                "a4810000"  # st_uid
                "00000000"  # st_gid
                "00000000"
                "00000000"  # st_rdev
                "00000000"
                "00000000",
            ),  # __pad1
            int.to_bytes(stat_result.st_size, 8, "little"),  # st_size
            bytes.fromhex(
                "00000000"  # st_blksize
                "00000000"  # __pad2
                "00000000"
                "00000000"  # st_blocks
                "00000000"
                "00000000"  # st_atime
                "00000000"
                "00000000" +  # st_atime_nsec
                # "00" * 4 +
                "00" * 2
                + "01" * 2
                + "00000000"  # st_mtime [This must have a valid value]
                "00000000"
                "00000000"  # st_mtime_nsec
                "00000000"
                "00000000"  # st_ctime
                "00000000"
                "00000000"  # st_ctime_nsec
                "00000000"  # __unused4
                "00000000",  # __unused5
            ),
        ],
    )
    logger.debug(len(stat_bytes))
    assert len(stat_bytes) in [104, 128]

    ctx.vm.mem_write(buf, stat_bytes)

    # Return success
    ctx.vm.reg_write(UC_ARM64_REG_X0, 0)


def _hook_lstat(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)

    p_path = x0
    path = ctx.vm.read_cstr(p_path).decode("utf-8")
    buf = x1

    logger.debug("lstat(0x%X:'%s', [...])", p_path, path)

    return _handle_stat(ctx, path, buf)


def _hook_fstat(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)

    fildes = x0
    buf = x1

    return _handle_stat(ctx, fildes, buf)


def _hook_open(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)
    x2 = ctx.vm.reg_read(UC_ARM64_REG_X2)

    path = ctx.vm.read_cstr(x0).decode("utf-8")
    oflag = x1
    mode = x2

    logger.debug("open('%s', %s, %s)", path, oct(oflag), oct(mode))
    # time.sleep(2.0)
    # assert(False)

    # Return fildes
    fildes = ctx.fs.open(path, oflag)
    ctx.vm.reg_write(UC_ARM64_REG_X0, fildes)


def _hook_ftruncate(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)

    fildes = x0
    length = x1

    logger.debug("ftruncate(%d, %d)", fildes, length)

    ctx.fs.truncate(fildes, length)

    ctx.vm.reg_write(UC_ARM64_REG_X0, 0)


def _hook_read(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)
    x2 = ctx.vm.reg_read(UC_ARM64_REG_X2)

    fildes = x0
    buf = x1
    nbyte = x2

    logger.debug("read(%d, 0x%X, %d)", fildes, buf, nbyte)

    buf_bytes = ctx.fs.read(fildes, nbyte)
    ctx.vm.mem_write(buf, buf_bytes)

    ctx.vm.reg_write(UC_ARM64_REG_X0, nbyte)


def _hook_write(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)
    x2 = ctx.vm.reg_read(UC_ARM64_REG_X2)

    fildes = x0
    buf = x1
    nbyte = x2

    logger.debug("write(%d, 0x%X, %d)", fildes, buf, nbyte)

    buf_bytes = ctx.vm.mem_read(buf, nbyte)
    ctx.fs.write(fildes, buf_bytes)

    ctx.vm.reg_write(UC_ARM64_REG_X0, nbyte)


def _hook_close(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)

    fildes = x0

    ctx.fs.close(fildes)

    ctx.vm.reg_write(UC_ARM64_REG_X0, 0)


def _hook_dlopen_wrapper(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    path = ctx.vm.read_cstr(x0).decode("utf-8")
    library_name = path.rpartition("/")[2]

    logger.debug("dlopen('%s' (%s))", path, library_name)

    assert library_name in [
        "libCoreADI.so",
    ]

    library = ctx.vm.load_library(library_name)
    x0 = library.index
    ctx.vm.reg_write(UC_ARM64_REG_X0, 1 + x0)

    # assert(False)


def _hook_dlsym_wrapper(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)
    handle = x0
    symbol = ctx.vm.read_cstr(x1).decode("utf-8")

    library_index = handle - 1
    library = ctx.vm.get_library(library_index)

    logger.debug("dlsym(%X (%s), '%s')", handle, library.name, symbol)

    symbol_address = library.resolve_symbol_by_name(symbol)
    logger.debug("Found at 0x%X", symbol_address)

    ctx.vm.reg_write(UC_ARM64_REG_X0, symbol_address)


def _hook_gettimeofday(ctx: HookContext) -> None:
    timestamp = time.time()

    # FIXME: why need caching??
    cache_time = False
    cache_path = "cache/time.bin"

    if cache_time:
        t_bytes = open(cache_path, "rb").read()  # noqa: PTH123, SIM115
        logger.debug("Loaded time from cache!")
        t = c_timeval.from_buffer_copy(t_bytes)
        logger.debug("ok: %s", t)

    t = c_timeval(
        tv_sec=math.floor(timestamp // 1),
        tv_usec=math.floor((timestamp % 1.0) * 1000 * 1000),
    )
    t_bytes = bytes(t)

    if cache_time:
        open(cache_path, "wb").write(t_bytes)  # noqa: PTH123, SIM115

    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)

    tp = x0
    tzp = x1

    logger.debug("gettimeofday(0x%X, 0x%X)", tp, tzp)

    # We don't need timezone support
    assert tzp == 0
    """
    struct timezone {
             int     tz_minuteswest; /* of Greenwich */
             int     tz_dsttime;     /* type of dst correction to apply */
    };
    """

    # Write the time
    logger.debug("%s %s %s", t.__dict__, t_bytes.hex(), len(t_bytes))
    ctx.vm.mem_write(tp, t_bytes)

    # Return success
    ctx.vm.reg_write(UC_ARM64_REG_X0, 0)


def _hook_errno_location(ctx: HookContext) -> None:
    if ctx.vm.errno_address is None:
        logger.debug("Checking errno before first error (!)")
        ctx.vm.set_errno(0)
        assert ctx.vm.errno_address is not None

    ctx.vm.reg_write(UC_ARM64_REG_X0, ctx.vm.errno_address)


def _hook_system_property_get_impl(ctx: HookContext) -> None:
    x0 = ctx.vm.reg_read(UC_ARM64_REG_X0)
    x1 = ctx.vm.reg_read(UC_ARM64_REG_X1)
    name = ctx.vm.read_cstr(x0).decode("utf-8")
    logger.debug("__system_property_get(%s, [...])", name)
    value = b"no s/n number"
    ctx.vm.mem_write(x1, value)
    ctx.vm.reg_write(UC_ARM64_REG_X0, len(value))


def _hook_arc4random_impl(ctx: HookContext) -> None:
    value = 0xDEADBEEF  # "Random number, chosen by fair dice roll"
    ctx.vm.reg_write(UC_ARM64_REG_X0, value)


STUBBED_FUNCTIONS = {
    # memory management
    "malloc": _hook_malloc,
    "free": _hook_free,
    # string
    "strncpy": _hook_strncpy,
    # fs
    "mkdir": _hook_mkdir,
    "umask": _hook_umask,
    "chmod": _hook_chmod,
    "lstat": _hook_lstat,
    "fstat": _hook_fstat,
    # io
    "open": _hook_open,
    "ftruncate": _hook_ftruncate,
    "read": _hook_read,
    "write": _hook_write,
    "close": _hook_close,
    # dynamic symbol stuff
    "dlsym": _hook_dlsym_wrapper,
    "dlopen": _hook_dlopen_wrapper,
    "dlclose": lambda ctx: _hook_empty_stub(ctx, "dlclose"),
    # pthreads
    "pthread_once": lambda ctx: _hook_empty_stub(ctx, "pthread_once"),
    "pthread_create": lambda ctx: _hook_empty_stub(ctx, "pthread_create"),
    "pthread_mutex_lock": lambda ctx: _hook_empty_stub(ctx, "pthread_mutex_lock"),
    "pthread_rwlock_unlock": lambda ctx: _hook_empty_stub(ctx, "pthread_rwlock_unlock"),
    "pthread_rwlock_destroy": lambda ctx: _hook_empty_stub(ctx, "pthread_rwlock_destroy"),
    "pthread_rwlock_wrlock": lambda ctx: _hook_empty_stub(ctx, "pthread_rwlock_wrlock"),
    "pthread_rwlock_init": lambda ctx: _hook_empty_stub(ctx, "pthread_rwlock_init"),
    "pthread_mutex_unlock": lambda ctx: _hook_empty_stub(ctx, "pthread_mutex_unlock"),
    "pthread_rwlock_rdlock": lambda ctx: _hook_empty_stub(ctx, "pthread_rwlock_rdlock"),
    # date and time
    "gettimeofday": _hook_gettimeofday,
    # misc
    "__errno": _hook_errno_location,
    "__system_property_get": _hook_system_property_get_impl,
    "arc4random": _hook_arc4random_impl,
}


def hook_mem_invalid(_ctx: HookContext, access: int, address: int, size: int, value: int) -> None:
    if access == UC_MEM_WRITE_UNMAPPED:
        logger.error(
            ">>> Missing memory is being WRITE at 0x%x, data size = %u, data value = 0x%x",
            address,
            size,
            value,
        )
        # return True to indicate we want to continue emulation
        # return False
    elif access == UC_MEM_FETCH_UNMAPPED:
        logger.error(
            ">>> Missing memory is being FETCH at 0x%x, data size = %u, data value = 0x%x",
            address,
            size,
            value,
        )
    else:
        # return False to indicate we want to stop emulation
        # return False
        msg = f"Unsupported access mode: {access}"
        raise RuntimeError(msg)


def hook_code(ctx: HookContext, address: int, size: int) -> None:
    logs = ""
    logs += f">>> Tracing at 0x{address:X}:"
    # read this instruction code from memory
    for i in ctx.vm.mem_read(address, size):
        logs += f" {i:02X}"
    for i in [3, 8, 9, 10, 11, 20]:
        value = ctx.vm.reg_read(UC_ARM64_REG_X0 + i)
        logs += f"; X{i:d}: 0x{value:08X}"
    logs += f"; W13=0x{ctx.vm.reg_read(UC_ARM64_REG_W13):X}"
    logs += f"; W14=0x{ctx.vm.reg_read(UC_ARM64_REG_W14):X}"
    logs += f"; W15=0x{ctx.vm.reg_read(UC_ARM64_REG_W15):X}"
    logs += f"; FP/X29=0x{ctx.vm.reg_read(UC_ARM64_REG_FP):X}"
    # print("; *347c40=0x%08X" % int.from_bytes(uc.mem_read(0x347c40, 4), 'little'), end="")
    logger.debug(logs)


def hook_block(ctx: HookContext, _address: int, _size: int) -> None:
    pass
    # print("         >>> Tracing basic block at 0x%x, block size = 0x%x" %(address, size))


def hook_stub(ctx: HookContext, address: int, _size: int) -> bool:
    assert address >= IMPORT_ADDRESS
    assert address < IMPORT_ADDRESS + 0x01000000 * 10

    offset = address - IMPORT_ADDRESS
    library_index = offset // 0x01000000
    symbol_index = (offset % 0x01000000) // 4

    # assert(libraryIndex == 0)
    library = ctx.vm.get_library(library_index)

    symbol_name = library.symbol_name_by_index(symbol_index)

    # lr = uc.reg_read(UC_ARM64_REG_LR)

    # print("stub", "0x%X" % lr, uc, address, size, user_data, end=" :: ")
    # print(libraryIndex, library.name, symbolIndex, symbolName)

    if symbol_name in STUBBED_FUNCTIONS:
        STUBBED_FUNCTIONS[symbol_name](ctx)
        # assert(False)
    else:
        msg = f"Symbol not in stubbed functions: {symbol_name}"
        raise RuntimeError(msg)

    # time.sleep(0.1)
    return True



from __future__ import annotations

from anisette import Anisette


def test_init_save():
    ani = Anisette.init("applemusic.apk")

    assert isinstance(ani.get_data(), dict)

    ani.save_all("bundle.bin")


def test_load():
    ani = Anisette.load("bundle.bin")

    assert isinstance(ani.get_data(), dict)



from __future__ import annotations

import io
import logging
import tarfile
import zipfile
from typing import IO, TYPE_CHECKING, BinaryIO

from elftools.elf.sections import SymbolTableSection
from typing_extensions import Self

from ._arch import Architecture
from ._fs import VirtualFileSystem

if TYPE_CHECKING:
    from elftools.elf.elffile import ELFFile


logging.getLogger(__name__)

R_AARCH64_ABS64 = 257
R_AARCH64_GLOB_DAT = 1025
R_AARCH64_JUMP_SLOT = 1026
R_AARCH64_RELATIVE = 1027


class Library:
    def __init__(self, name: str, elf: ELFFile, base: int, index: int) -> None:
        self.name = name
        self.elf = elf
        self.base = base
        self.symbols = {}
        self.index = index

    def resolve_symbol_by_index(self, symbol_index: int) -> int:
        # for section in elf.iter_sections():
        #   print(section)
        if symbol_index in self.symbols:
            # print("Resolving symbol 0x%X from symbols dict" % symbolIndex)
            return self.symbols[symbol_index]

        section = self.elf.get_section_by_name(".dynsym")
        assert isinstance(section, SymbolTableSection)

        sym = section.get_symbol(symbol_index)
        # print("Resolving symbol 0x%X relative to base" % symbolIndex, sym.__dict__)

        # if sym['st_shndx'] == 11:
        #    section = library.elf.get_section(sym['st_shndx'])
        #    print("Fixing section", section.__dict__)
        #    print("0x%X" % (section['sh_addr'] + sym['st_value']))
        #    assert(False)

        return self.base + sym["st_value"]

    def resolve_symbol_by_name(self, symbol_name: str) -> int:
        section = self.elf.get_section_by_name(".dynsym")
        assert isinstance(section, SymbolTableSection)

        num_symbols = section.num_symbols()
        for i in range(num_symbols):
            sym = section.get_symbol(i)
            if sym.name == symbol_name:
                # print(sym.__dict__)
                return self.resolve_symbol_by_index(i)

        msg = f"Symbol '{symbol_name}' not found"
        raise ValueError(msg)

    def symbol_name_by_index(self, symbol_index: int) -> str:
        section = self.elf.get_section_by_name(".dynsym")
        assert isinstance(section, SymbolTableSection)

        sym = section.get_symbol(symbol_index)
        return sym.name


class LibraryStore(VirtualFileSystem):
    _LIBRARIES = (
        "libstoreservicescore.so",
        "libCoreADI.so",
    )
    _ARCH = Architecture.ARM64

    def __init__(self, fs: VirtualFileSystem | None) -> None:
        super().__init__(fs)

    def open_library(self, name: str) -> IO:
        return self.easy_open(name, "rb")

    def add_library(self, name: str, data: IO[bytes]) -> None:
        with self.easy_open(name, "wb+") as f:
            f.write(data.read())

    @staticmethod
    def _candidates_for(lib: str, arch: Architecture) -> tuple[str, str, str]:
        return (
            lib,
            f"libs/{lib}",
            f"lib/{arch.value}/{lib}",
        )

    @classmethod
    def _load_from_tar(cls, f: BinaryIO, lib_store: LibraryStore) -> bool:
        try:
            with tarfile.open(fileobj=f, mode="r:*") as tf:
                names = {m.name for m in tf.getmembers() if m.isfile()}
                for lib in cls._LIBRARIES:
                    for path in cls._candidates_for(lib, cls._ARCH):
                        if path in names:
                            member = tf.getmember(path)
                            data = tf.extractfile(member)
                            if data is None:
                                continue
                            with data:
                                lib_store.add_library(lib, data)
                            break
                    else:
                        msg = "Archive is missing library file: %s"
                        raise RuntimeError(msg % lib)
        except tarfile.ReadError:
            return False
        else:
            return True

    @classmethod
    def _load_from_zip(cls, f: BinaryIO, lib_store: LibraryStore) -> bool:
        try:
            with zipfile.ZipFile(f) as zf:
                names = set(zf.namelist())
                for lib in cls._LIBRARIES:
                    for path in cls._candidates_for(lib, cls._ARCH):
                        if path in names:
                            with zf.open(path, "r") as data:
                                lib_store.add_library(lib, data)
                            break
                    else:
                        msg = "Archive is missing library file: %s"
                        raise RuntimeError(msg % lib)
        except zipfile.BadZipFile:
            return False
        else:
            return True

    @classmethod
    def from_file(cls, file: BinaryIO) -> Self:
        """Load libraries from a tar or zip archive using only stdlib."""
        lib_store = cls(None)

        # Buffer the file so we can attempt multiple formats without relying on seekability.
        data = file.read()
        buf1 = io.BytesIO(data)
        if cls._load_from_tar(buf1, lib_store):
            return lib_store

        buf2 = io.BytesIO(data)
        if cls._load_from_zip(buf2, lib_store):
            return lib_store

        msg = "Unknown file format"
        raise TypeError(msg)



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from ._util import u_to_s32
from ._vm import VM, Architecture

if TYPE_CHECKING:
    from ._fs import VirtualFileSystem
    from ._library import LibraryStore

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ClientProvisioningIntermediateMetadata:
    adi: ADI
    cpim: bytes
    session: int


@dataclass(frozen=True)
class OneTimePassword:
    adi: ADI
    otp: bytes
    machine_id: bytes


class ADI:
    def __init__(self, fs: VirtualFileSystem, lib_store: LibraryStore, identifier: str) -> None:
        self._vm = VM.create(fs, lib_store, Architecture.ARM64)

        self._provisioning_path: str | None = None
        self._identifier: str | None = None

        ssc_library = self._vm.load_library("libstoreservicescore.so")

        logger.debug("Loading Android-specific symbols...")

        self.__pADILoadLibraryWithPath = ssc_library.resolve_symbol_by_name("kq56gsgHG6")
        self.__pADISetAndroidID = ssc_library.resolve_symbol_by_name("Sph98paBcz")
        self.__pADISetProvisioningPath = ssc_library.resolve_symbol_by_name("nf92ngaK92")

        logger.debug("Loading ADI symbols...")

        self.__pADIProvisioningErase = ssc_library.resolve_symbol_by_name("p435tmhbla")
        self.__pADISynchronize = ssc_library.resolve_symbol_by_name("tn46gtiuhw")
        self.__pADIProvisioningDestroy = ssc_library.resolve_symbol_by_name("fy34trz2st")
        self.__pADIProvisioningEnd = ssc_library.resolve_symbol_by_name("uv5t6nhkui")
        self.__pADIProvisioningStart = ssc_library.resolve_symbol_by_name("rsegvyrt87")
        self.__pADIGetLoginCode = ssc_library.resolve_symbol_by_name("aslgmuibau")
        self.__pADIDispose = ssc_library.resolve_symbol_by_name("jk24uiwqrg")
        self.__pADIOTPRequest = ssc_library.resolve_symbol_by_name("qi864985u0")

        self._set_identifier(identifier)
        self._set_provisioning_path(".")
        self._load_library(".")

    @property
    def alloc_stats(self) -> tuple[float, float, float]:
        return self._vm.alloc_stats

    def _set_provisioning_path(self, value: str) -> None:
        p_path = self._vm.temp_alloc_data(value.encode("utf-8") + b"\x00")
        self._vm.invoke_cdecl(self.__pADISetProvisioningPath, [p_path])
        self._provisioning_path = value
        self._vm.temp_free(p_path)

    def _set_identifier(self, value: str) -> None:
        self._identifier = value
        logger.debug("Setting identifier %s", value)
        identifier = value.encode("utf-8")
        p_identifier = self._vm.temp_alloc_data(identifier)
        self._vm.invoke_cdecl(self.__pADISetAndroidID, [p_identifier, len(identifier)])
        self._vm.temp_free(p_identifier)

    def _load_library(self, library_path: str) -> None:
        p_library_path = self._vm.temp_alloc_data(library_path.encode("utf-8") + b"\x00")
        self._vm.invoke_cdecl(self.__pADILoadLibraryWithPath, [p_library_path])
        self._vm.temp_free(p_library_path)

    def erase_provisioning(self) -> None:
        raise NotImplementedError

    def synchronize(self) -> None:
        raise NotImplementedError

    def destroy_provisioning(self) -> None:
        raise NotImplementedError

    def end_provisioning(self, session: int, persistent_token_metadata: bytes, trust_key: bytes) -> None:
        p_persistent_token_metadata = self._vm.temp_alloc_data(persistent_token_metadata)
        p_trust_key = self._vm.temp_alloc_data(trust_key)

        ret = self._vm.invoke_cdecl(
            self.__pADIProvisioningEnd,
            [
                session,
                p_persistent_token_metadata,
                len(persistent_token_metadata),
                p_trust_key,
                len(trust_key),
            ],
        )

        self._vm.temp_free(p_persistent_token_metadata)
        self._vm.temp_free(p_trust_key)

        logger.debug("0x%X", session)
        logger.debug("Persistent token: %s (len: %i)", persistent_token_metadata.hex(), len(persistent_token_metadata))
        logger.debug("Trust key: %s (len: %d)", trust_key.hex(), len(trust_key))

        logger.debug("%s: %X=%d", "pADIProvisioningEnd", ret, u_to_s32(ret))
        assert ret == 0

    def start_provisioning(
        self,
        ds_id: int,
        server_provisioning_intermediate_metadata: bytes,
    ) -> ClientProvisioningIntermediateMetadata:
        logger.debug("ADI.start_provisioning")
        # FIXME: !!!

        p_cpim = self._vm.temp_alloc(8)  # ubyte*
        p_cpim_length = self._vm.temp_alloc(4)  # uint
        p_session = self._vm.temp_alloc(4)  # uint
        p_server_provisioning_intermediate_metadata = self._vm.temp_alloc_data(
            server_provisioning_intermediate_metadata,
        )
        logger.debug("0x%X", ds_id)
        logger.debug(server_provisioning_intermediate_metadata.hex())

        ret = self._vm.invoke_cdecl(
            self.__pADIProvisioningStart,
            [
                ds_id,
                p_server_provisioning_intermediate_metadata,
                len(server_provisioning_intermediate_metadata),
                p_cpim,
                p_cpim_length,
                p_session,
            ],
        )
        logger.debug("%s: %X=%d", "pADIProvisioningStart", ret, u_to_s32(ret))
        assert ret == 0

        self._vm.temp_free(p_cpim)
        self._vm.temp_free(p_cpim_length)
        self._vm.temp_free(p_session)
        self._vm.temp_free(p_server_provisioning_intermediate_metadata)

        # Readback output
        cpim = self._vm.read_u64(p_cpim)
        logger.debug("Wrote data to 0x%X", cpim)
        cpim_length = self._vm.read_u32(p_cpim_length)
        cpim_bytes = self._vm.mem_read(cpim, cpim_length)
        session = self._vm.read_u32(p_session)

        # logger.debug(cpim_length, cpim_bytes.hex(), session)
        # assert(False)
        return ClientProvisioningIntermediateMetadata(self, cpim_bytes, session)

    def is_machine_provisioned(self, ds_id: int) -> bool:
        logger.debug("ADI.is_machine_provisioned")

        error_code = u_to_s32(self._vm.invoke_cdecl(self.__pADIGetLoginCode, [ds_id]))

        if error_code == 0:
            return True
        if error_code == -45061:
            return False

        msg = f"Unknown errorCode: {error_code:d}=0x{error_code:X}"
        raise RuntimeError(msg)

    def dispose(self) -> None:
        raise NotImplementedError

    def request_otp(self, ds_id: int) -> OneTimePassword:
        logger.debug("ADI.request_otp")
        # FIXME: !!!

        p_otp = self._vm.temp_alloc(8)
        p_otp_length = self._vm.temp_alloc(4)
        p_mid = self._vm.temp_alloc(8)
        p_mid_length = self._vm.temp_alloc(4)

        # ubyte* otp;
        # uint otpLength;
        # ubyte* mid;
        # uint midLength;

        ret = self._vm.invoke_cdecl(
            self.__pADIOTPRequest,
            [
                ds_id,
                p_mid,
                p_mid_length,
                p_otp,
                p_otp_length,
            ],
        )
        logger.debug("%s: %X=%d", "pADIOTPRequest", ret, u_to_s32(ret))
        assert ret == 0

        self._vm.temp_free(p_otp)
        self._vm.temp_free(p_otp_length)
        self._vm.temp_free(p_mid)
        self._vm.temp_free(p_mid_length)

        otp = self._vm.read_u64(p_otp)
        otp_length = self._vm.read_u32(p_otp_length)
        otp_bytes = self._vm.mem_read(otp, otp_length)

        mid = self._vm.read_u64(p_mid)
        mid_length = self._vm.read_u32(p_mid_length)
        mid_bytes = self._vm.mem_read(mid, mid_length)

        return OneTimePassword(self, otp_bytes, mid_bytes)



from __future__ import annotations

import json
import secrets
import uuid
from dataclasses import dataclass
from typing import TYPE_CHECKING

from typing_extensions import Self

if TYPE_CHECKING:
    from ._fs import VirtualFileSystem


@dataclass()
class AnisetteDeviceConfig:
    server_friendly_description: str
    unique_device_id: str
    adi_id: str
    local_user_uuid: str

    @classmethod
    def default(cls) -> Self:
        return cls(
            server_friendly_description=(
                "<MacBookPro13,2> <macOS;13.1;22C65> <com.apple.AuthKit/1 (com.apple.dt.Xcode/3594.4.19)>"
            ),
            unique_device_id=str(uuid.uuid4()).upper(),
            adi_id=secrets.token_hex(8).lower(),
            local_user_uuid=secrets.token_hex(32).upper(),
        )


class Device:
    _UNIQUE_DEVICE_IDENTIFIER_JSON = "UUID"
    _SERVER_FRIENDLY_DESCRIPTION_JSON = "clientInfo"
    _ADI_IDENTIFIER_JSON = "identifier"
    _LOCAL_USER_UUID_JSON = "localUUID"

    _PATH = "device.json"

    def __init__(self, fs: VirtualFileSystem, default_config: AnisetteDeviceConfig) -> None:
        self._fs = fs

        # Attempt to load the JSON
        try:
            with self._fs.easy_open(self._PATH, "r") as f:
                data = json.load(f)
        except FileNotFoundError:
            data = {}

        self._unique_device_identifier: str = data.get(
            self._UNIQUE_DEVICE_IDENTIFIER_JSON,
            default_config.unique_device_id,
        )
        self._server_friendly_description: str = data.get(
            self._SERVER_FRIENDLY_DESCRIPTION_JSON,
            default_config.server_friendly_description,
        )
        self._adi_identifier = data.get(
            self._ADI_IDENTIFIER_JSON,
            default_config.adi_id,
        )
        self._local_user_uuid = data.get(
            self._LOCAL_USER_UUID_JSON,
            default_config.local_user_uuid,
        )

        self.write()

    def write(self) -> None:
        # Save to JSON
        data = {
            self._UNIQUE_DEVICE_IDENTIFIER_JSON: self._unique_device_identifier,
            self._SERVER_FRIENDLY_DESCRIPTION_JSON: self._server_friendly_description,
            self._ADI_IDENTIFIER_JSON: self._adi_identifier,
            self._LOCAL_USER_UUID_JSON: self._local_user_uuid,
        }
        with self._fs.easy_open(self._PATH, "w") as f:
            json.dump(data, f)

    @property
    def unique_device_identifier(self) -> str:
        return self._unique_device_identifier

    @property
    def server_friendly_description(self) -> str:
        return self._server_friendly_description

    @property
    def adi_identifier(self) -> str:
        return self._adi_identifier

    @property
    def local_user_uuid(self) -> str:
        return self._local_user_uuid



from ctypes import Structure, c_int, c_long, c_size_t, c_uint32, c_uint64, c_ulong

# Based on https://github.com/Dadoum/Provision/blob/main/lib/std_edit/linux_stat.d (aarch64)
# FIXME: These must be changed to fixed size types
c_dev_t = c_uint32
c_off_t = c_size_t
c_ino_t = c_uint64
c_mode_t = c_uint32  # c_ushort
c_nlink_t = c_uint32
c_uid_t = c_uint32
c_gid_t = c_uint32
c_blksize_t = c_ulong
c_blkcnt_t = c_uint64
c_time_t = c_uint64
c_suseconds_t = c_long


class c_timeval(Structure):  # noqa: N801
    _fields_ = (
        ("tv_sec", c_time_t),  # /* seconds since Jan. 1, 1970 */
        ("tv_usec", c_suseconds_t),  # /* and microseconds */
    )


class c_stat(Structure):  # noqa: N801
    _fields_ = (
        ("st_dev", c_dev_t),  # /* ID of device containing file */
        ("st_ino", c_ino_t),  # /* inode number */
        ("st_mode", c_mode_t),  # /* protection */
        ("st_nlink", c_nlink_t),  # /* number of hard links */
        ("st_uid", c_uid_t),  # /* user ID of owner */
        ("st_gid", c_gid_t),  # /* group ID of owner */
        ("st_rdev", c_dev_t),  # /* device ID (if special file) */
        ("__pad1", c_dev_t),  # ???
        ("st_size", c_off_t),  # /* total size, in bytes */
        ("st_blksize", c_blksize_t),  # /* blocksize for file system I/O */
        ("__pad2", c_int),  # ???
        ("st_blocks", c_blkcnt_t),  # /* number of 512B blocks allocated */
        ("st_atime", c_time_t),  # /* time of last access */
        ("st_atimensec", c_ulong),  # ?!?!
        ("st_mtime", c_time_t),  # /* time of last modification */
        ("st_mtimensec", c_ulong),  # ?!?!
        ("st_ctime", c_time_t),  # /* time of last status change */
        ("st_ctimensec", c_ulong),  # ?!?!
        ("__unused_0", c_int),  # ???
        ("__unused_1", c_int),  # ???
    )


# From https://chromium.googlesource.com/android_tools/+/20ee6d20/ndk/platforms/android-21/arch-arm64/usr/include/sys/stat.h
"""
  unsigned long st_dev; \
  unsigned long st_ino; \
  unsigned int st_mode; \
  unsigned int st_nlink; \
  uid_t st_uid; \
  gid_t st_gid; \
  unsigned long st_rdev; \
  unsigned long __pad1; \
  long st_size; \
  int st_blksize; \
  int __pad2; \
  long st_blocks; \
  long st_atime; \
  unsigned long st_atime_nsec; \
  long st_mtime; \
  unsigned long st_mtime_nsec; \
  long st_ctime; \
  unsigned long st_ctime_nsec; \
  unsigned int __unused4; \
  unsigned int __unused5; \
"""



from __future__ import annotations

import io as _io
import logging
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable

from elftools.elf.elffile import ELFFile
from elftools.elf.relocation import RelocationSection
from elftools.elf.sections import SymbolTableSection
from unicorn import (
    UC_ARCH_ARM,
    UC_ARCH_ARM64,
    UC_ARCH_X86,
    UC_HOOK_BLOCK,
    UC_HOOK_CODE,
    UC_HOOK_MEM_FETCH_UNMAPPED,
    UC_HOOK_MEM_READ_UNMAPPED,
    UC_HOOK_MEM_WRITE_UNMAPPED,
    UC_MODE_32,
    UC_MODE_64,
    UC_MODE_ARM,
)
from unicorn.arm64_const import (
    UC_ARM64_REG_LR,
    UC_ARM64_REG_SP,
    UC_ARM64_REG_X0,
)
from unicorn.unicorn import Uc

from ._allocator import Allocator
from ._arch import Architecture
from ._hooks import HookContext, hook_block, hook_mem_invalid, hook_stub
from ._library import (
    R_AARCH64_ABS64,
    R_AARCH64_GLOB_DAT,
    R_AARCH64_JUMP_SLOT,
    R_AARCH64_RELATIVE,
    Library,
    LibraryStore,
)

if TYPE_CHECKING:
    from ._fs import VirtualFileSystem

logger = logging.getLogger(__name__)

RETURN_ADDRESS = 0xDEAD0000
STACK_ADDRESS = 0xF0000000
STACK_SIZE = 0x100000

MALLOC_ADDRESS = 0x60000000
MALLOC_SIZE = 0x1000000

TEMP_ADDRESS = 0x800000000
TEMP_SIZE = 0x100000

IMPORT_ADDRESS = 0xA0000000
IMPORT_SIZE = 0x1000


class VM:
    def __init__(self, uc: Uc, fs: VirtualFileSystem, lib_store: LibraryStore) -> None:
        self._uc = uc
        self._fs = fs

        self._lib_store = lib_store
        self._loaded_libs: dict[str, Library] = OrderedDict()

        self._temp_allocator = Allocator(TEMP_ADDRESS, TEMP_SIZE)
        self._malloc_allocator = Allocator(MALLOC_ADDRESS, MALLOC_SIZE)
        self._lib_allocator = Allocator(0x00100000, 0x90000000)

        self._errno_address: int | None = None

    @property
    def alloc_stats(self) -> tuple[float, float, float]:
        return (
            self._temp_allocator.alloc_perc,
            self._malloc_allocator.alloc_perc,
            self._lib_allocator.alloc_perc,
        )

    @property
    def errno_address(self) -> int | None:
        return self._errno_address

    def wrap_hook(self, hook: Callable) -> Callable:
        def _new_hook(_uc: Uc, *args: Any) -> None:  # noqa: ANN401
            ctx = HookContext(vm=self, fs=self._fs)
            return hook(ctx, *(args[:-1]))

        return _new_hook

    @classmethod
    def create(cls, fs: VirtualFileSystem, lib_store: LibraryStore, arch: Architecture) -> VM:
        # Startup a unicorn-engine instance as VM backend
        if arch == Architecture.X86:
            uc = Uc(UC_ARCH_X86, UC_MODE_32)
        elif arch == Architecture.X86_64:
            uc = Uc(UC_ARCH_X86, UC_MODE_64)
        elif arch == Architecture.ARM:
            uc = Uc(UC_ARCH_ARM, UC_MODE_ARM)
        elif arch == Architecture.ARM64:
            uc = Uc(UC_ARCH_ARM64, UC_MODE_ARM)
        else:
            msg = "Invalid architecture: %s"
            raise ValueError(msg, arch)

        # Register a fake return address
        uc.mem_map(RETURN_ADDRESS, 0x1000)

        # Register some memory for malloc
        uc.mem_map(MALLOC_ADDRESS, MALLOC_SIZE)

        # Register memory for temp data
        uc.mem_map(TEMP_ADDRESS, TEMP_SIZE)

        # Register a fake stack
        uc.mem_map(STACK_ADDRESS, STACK_SIZE)

        vm = cls(uc, fs, lib_store)

        # Debug hooks
        uc.hook_add(UC_HOOK_BLOCK, vm.wrap_hook(hook_block))
        # uc.hook_add(UC_HOOK_CODE, hook_code, vm)
        uc.hook_add(
            UC_HOOK_MEM_READ_UNMAPPED | UC_HOOK_MEM_WRITE_UNMAPPED | UC_HOOK_MEM_FETCH_UNMAPPED,
            vm.wrap_hook(hook_mem_invalid),
        )

        # Add a region for imports
        import_count = IMPORT_SIZE // 4
        for i in range(10):
            library_import_address = IMPORT_ADDRESS + i * 0x01000000
            uc.mem_map(library_import_address, IMPORT_SIZE)
            uc.mem_write(library_import_address, b"\xc0\x03\x5f\xd6" * import_count)  # RET instruction
            uc.hook_add(
                UC_HOOK_CODE,
                vm.wrap_hook(hook_stub),
                None,
                library_import_address,
                library_import_address + IMPORT_SIZE - 1,
            )

        return vm

    def malloc(self, length: int) -> int:
        return self._malloc_allocator.alloc(length)[0]

    def free(self, address: int) -> None:
        return self._malloc_allocator.free(address)

    def mem_write(self, address: int, data: bytes) -> None:
        self._uc.mem_write(address, data)

    def mem_read(self, address: int, length: int) -> bytes:
        return bytes(self._uc.mem_read(address, length))

    def reg_write(self, reg_id: int, value: int) -> None:
        return self._uc.reg_write(reg_id, value)

    def reg_read(self, reg_id: int) -> int:
        return self._uc.reg_read(reg_id)

    def write_u64(self, address: int, value: int) -> None:
        return self.mem_write(
            address,
            int.to_bytes(value, 8, "little", signed=False),
        )

    def read_u64(self, address: int) -> int:
        return int.from_bytes(
            self.mem_read(address, 8),
            "little",
            signed=False,
        )

    def write_u32(self, address: int, value: int) -> None:
        return self.mem_write(
            address,
            int.to_bytes(value, 4, "little", signed=False),
        )

    def read_u32(self, address: int) -> int:
        return int.from_bytes(
            self.mem_read(address, 4),
            "little",
            signed=False,
        )

    def read_cstr(self, address: int) -> bytes:
        max_length = 0x1000
        s = self.mem_read(address, max_length)
        s, terminator, _ = s.partition(b"\x00")
        assert terminator == b"\x00"
        return s

    def set_errno(self, value: int) -> None:
        if self._errno_address is None:
            self._errno_address = self.temp_alloc(4)
        self.write_u32(self._errno_address, value)

    def temp_alloc_data(self, data: bytes) -> int:
        data_size = len(data)
        address, alloc_size = self._temp_allocator.alloc(data_size + 1)

        logger.debug("Allocating at 0x%X; bytes 0x%X/0x%X", address, data_size, alloc_size)
        self.mem_write(address, data + b"\xcc" * alloc_size)

        return address

    def temp_alloc(self, size: int) -> int:
        return self.temp_alloc_data(b"\xaa" * size)

    def temp_free(self, address: int) -> None:
        return self._temp_allocator.free(address)

    def invoke_cdecl(self, address: int, args: list[int]) -> int:
        lr = RETURN_ADDRESS
        for i, value in enumerate(args):
            assert i <= 28
            self.reg_write(UC_ARM64_REG_X0 + i, value)
            logger.debug("X%d: 0x%08X", i, value)
        logger.debug("Calling 0x%X", address)
        self.reg_write(UC_ARM64_REG_SP, STACK_ADDRESS + STACK_SIZE)
        self.reg_write(UC_ARM64_REG_LR, lr)
        # uc.reg_write(UC_ARM64_REG_FP, stackAddress + stackSize)
        self._uc.emu_start(address, lr)
        return self.reg_read(UC_ARM64_REG_X0)

    def relocate_section(self, library: Library, section_name: str, base: int) -> None:
        reladyn = library.elf.get_section_by_name(section_name)
        assert isinstance(reladyn, RelocationSection)

        for reloc in reladyn.iter_relocations():
            # print('    Relocation (%s)' % 'RELA' if reloc.is_RELA() else 'REL', end="")
            # Relocation entry attributes are available through item lookup
            # print('      offset = 0x%X' % reloc['r_offset'])
            # print("%s" % reloc.__dict__, end="")
            # print("")

            info_type = reloc["r_info_type"]
            address = base + reloc["r_offset"]

            if info_type in (R_AARCH64_ABS64, R_AARCH64_GLOB_DAT):
                symbol_index = reloc["r_info_sym"]
                symbol_address = library.resolve_symbol_by_index(symbol_index)
                self.mem_write(
                    address,
                    int.to_bytes(symbol_address + reloc["r_addend"], 8, "little"),
                )  # b'\x12\x34\x22\x78\xAB\xCD\xEF\xFF')
            elif info_type == R_AARCH64_JUMP_SLOT:
                symbol_index = reloc["r_info_sym"]
                symbol_address = library.resolve_symbol_by_index(symbol_index)
                self.mem_write(
                    address,
                    int.to_bytes(symbol_address, 8, "little"),
                )  # b'\x12\x34\x11\x78\xAB\xCD\xEF\xFF')
            elif info_type == R_AARCH64_RELATIVE:
                self.mem_write(
                    address,
                    int.to_bytes(base + reloc["r_addend"], 8, "little"),
                )  # b'\x12\x34\x22\x78\xAB\xCD\xEF\xFF')
            else:
                msg = "Invalid reloc info type: %d"
                raise RuntimeError(msg, info_type)

    def load_library(self, name: str) -> Library:
        if name in self._loaded_libs:
            return self._loaded_libs[name]

        library_index = len(self._loaded_libs)
        with self._lib_store.open_library(name) as f:
            elf_data = f.read()
            # Construct ELF from an in-memory buffer to avoid lifecycle issues of the context-managed stream
            elf = ELFFile(_io.BytesIO(elf_data))

        chosen_base = self._lib_allocator.alloc(0x10000000)[0]

        library = Library(name, elf, chosen_base, library_index)

        # Stub all imports
        section = library.elf.get_section_by_name(".dynsym")
        assert isinstance(section, SymbolTableSection)
        num_symbols = section.num_symbols()
        for i in range(num_symbols):
            sym = section.get_symbol(i)
            # print(sym.name)

            # print(sym.__dict__)
            # print(sym['st_shndx'])
            if sym["st_shndx"] == "SHN_UNDEF":
                library.symbols[i] = IMPORT_ADDRESS + library.index * 0x01000000 + i * 4
                # print("Registering 0x%X: %s" % (library.symbols[i], sym.name))

                # print("%s: 0x%X" % (sym.name, resolveSymbolByIndex(library, i)))

        for segment in library.elf.iter_segments():
            address = library.base + segment["p_vaddr"]
            size = segment["p_memsz"]

            address_start = address
            address_end = address + size

            alignment = segment["p_align"]

            # Align the start
            address_start &= ~(alignment - 1)

            # Align the end
            address_end += alignment - 1
            address_end &= ~(alignment - 1)

            # Fix size for new alignment
            size = address_end - address

            data_offset = segment["p_offset"]
            data_size = segment["p_filesz"]
            padding_before_size = address - address_start
            padding_after_size = size - data_size

            logger.debug(
                "Mapping at 0x%X-0x%X (0x%X-0x%X); bytes 0x%X",
                address_start,
                address_end,
                address,
                address + size - 1,
                size,
            )

            if segment["p_type"] == "PT_LOAD":
                data = (
                    b"\x00" * padding_before_size
                    + elf_data[data_offset : data_offset + data_size]
                    + b"\x00" * padding_after_size
                )
                self._uc.mem_map(address_start, len(data))
                self.mem_write(address_start, data)
            else:
                logger.debug("- Skipping %s", segment.__dict__)

        self.relocate_section(library, ".rela.dyn", library.base)
        self.relocate_section(library, ".rela.plt", library.base)

        self._loaded_libs[name] = library

        return library

    def get_library(self, index: int) -> Library:
        return list(self._loaded_libs.values())[index]
